@@ -766,13 +766,19 @@ static int type_bounds_copy_callback(hashtab_key_t key,
 	type_datum_t *type = (type_datum_t *) datum;
 	type_datum_t *dest;
 	uint32_t bounds_val;
+	hashtab_t type_tab;
 
 	if (!type->bounds)
 		return 0;
 
+	if (state->dest_decl == NULL)
+		type_tab = state->base->p_types.table;
+	else
+		type_tab = state->dest_decl->p_types.table;
+
 	bounds_val = state->cur->map[SYM_TYPES][type->bounds - 1];
 
-	dest = hashtab_search(state->base->p_types.table, key);
+	dest = hashtab_search(type_tab, key);
 	if (!dest) {
 		ERR(state->handle,
 		    "Type lookup failed for %s", (char *)key);
@@ -795,13 +801,19 @@ static int role_bounds_copy_callback(hashtab_key_t key,
 	role_datum_t *role = (role_datum_t *) datum;
 	role_datum_t *dest;
 	uint32_t bounds_val;
+	hashtab_t role_tab;
 
 	if (!role->bounds)
 		return 0;
 
+	if (state->dest_decl == NULL)
+		role_tab = state->base->p_roles.table;
+	else
+		role_tab = state->dest_decl->p_roles.table;
+
 	bounds_val = state->cur->map[SYM_ROLES][role->bounds - 1];
 
-	dest = hashtab_search(state->base->p_roles.table, key);
+	dest = hashtab_search(role_tab, key);
 	if (!dest) {
 		ERR(state->handle,
 		    "Role lookup failed for %s", (char *)key);
@@ -824,13 +836,19 @@ static int user_bounds_copy_callback(hashtab_key_t key,
 	user_datum_t *user = (user_datum_t *) datum;
 	user_datum_t *dest;
 	uint32_t bounds_val;
+	hashtab_t user_tab;
 
 	if (!user->bounds)
 		return 0;
 
+	if (state->dest_decl == NULL)
+		user_tab = state->base->p_users.table;
+	else
+		user_tab = state->dest_decl->p_users.table;
+
 	bounds_val = state->cur->map[SYM_USERS][user->bounds - 1];
 
-	dest = hashtab_search(state->base->p_users.table, key);
+	dest = hashtab_search(user_tab, key);
 	if (!dest) {
 		ERR(state->handle,
 		    "User lookup failed for %s", (char *)key);
@@ -1969,6 +1987,9 @@ static int is_decl_requires_met(link_state_t * state,
 
 
 		if (!is_id_enabled(id, state->base, SYM_CLASSES)) {
+			req->symbol_type = SYM_CLASSES;
+			req->symbol_value = i + 1;
+			req->perm_value = 0;
 			return 0;
 		}
 
@@ -2091,20 +2112,22 @@ static void print_missing_requirements(link_state_t * state,
 	    cur->branch_list->module_name : "BASE";
 
 	if (req->symbol_type == SYM_CLASSES) {
-
 		struct find_perm_arg fparg;
+		class_datum_t *cladatum = p->class_val_to_struct[req->symbol_value - 1];
 
-		class_datum_t *cladatum;
-		cladatum = p->class_val_to_struct[req->symbol_value - 1];
+		if (req->perm_value == 0) {
+			ERR(state->handle, "%s's global requirements were not met: class %s",
+				mod_name, p->p_class_val_to_name[req->symbol_value - 1]);
+		} else {
+			fparg.valuep = req->perm_value;
+			fparg.key = NULL;
+			(void)hashtab_map(cladatum->permissions.table, find_perm, &fparg);
 
-		fparg.valuep = req->perm_value;
-		fparg.key = NULL;
-		(void)hashtab_map(cladatum->permissions.table, find_perm, &fparg);
-
-		ERR(state->handle,
-		    "%s's global requirements were not met: class %s, permission %s",
-		    mod_name,
-		    p->p_class_val_to_name[req->symbol_value - 1], fparg.key);
+			ERR(state->handle,
+				"%s's global requirements were not met: class %s, permission %s",
+				mod_name,
+				p->p_class_val_to_name[req->symbol_value - 1], fparg.key);
+		}
 	} else {
 		ERR(state->handle,
 		    "%s's global requirements were not met: %s %s",
@@ -2348,120 +2371,6 @@ static int prepare_base(link_state_t * state, uint32_t num_mod_decls)
 	return 0;
 }
 
-static int expand_role_attributes(hashtab_key_t key, hashtab_datum_t datum,
-				  void * data)
-{
-	char *id;
-	role_datum_t *role, *sub_attr;
-	link_state_t *state;
-	unsigned int i;
-	ebitmap_node_t *rnode;
-
-	id = key;
-	role = (role_datum_t *)datum;
-	state = (link_state_t *)data;
-
-	if (strcmp(id, OBJECT_R) == 0){
-		/* object_r is never a role attribute by far */
-		return 0;
-	}
-
-	if (role->flavor != ROLE_ATTRIB)
-		return 0;
-
-	if (state->verbose)
-		INFO(state->handle, "expanding role attribute %s", id);
-
-restart:
-	ebitmap_for_each_positive_bit(&role->roles, rnode, i) {
-		sub_attr = state->base->role_val_to_struct[i];
-		if (sub_attr->flavor != ROLE_ATTRIB)
-			continue;
-
-		/* remove the sub role attribute from the parent
-		 * role attribute's roles ebitmap */
-		if (ebitmap_set_bit(&role->roles, i, 0))
-			return -1;
-
-		/* loop dependency of role attributes */
-		if (sub_attr->s.value == role->s.value)
-			continue;
-
-		/* now go on to expand a sub role attribute
-		 * by escalating its roles ebitmap */
-		if (ebitmap_union(&role->roles, &sub_attr->roles)) {
-			ERR(state->handle, "Out of memory!");
-			return -1;
-		}
-
-		/* sub_attr->roles may contain other role attributes,
-		 * re-scan the parent role attribute's roles ebitmap */
-		goto restart;
-	}
-
-	return 0;
-}
-
-/* For any role attribute in a declaration's local symtab[SYM_ROLES] table,
- * copy its roles ebitmap into its duplicate's in the base->p_roles.table.
- */
-static int populate_decl_roleattributes(hashtab_key_t key, 
-					hashtab_datum_t datum,
-					void *data)
-{
-	char *id = key;
-	role_datum_t *decl_role, *base_role;
-	link_state_t *state = (link_state_t *)data;
-
-	decl_role = (role_datum_t *)datum;
-
-	if (strcmp(id, OBJECT_R) == 0) {
-		/* object_r is never a role attribute by far */
-		return 0;
-	}
-
-	if (decl_role->flavor != ROLE_ATTRIB)
-		return 0;
-
-	base_role = (role_datum_t *)hashtab_search(state->base->p_roles.table,
-						   id);
-	assert(base_role != NULL && base_role->flavor == ROLE_ATTRIB);
-
-	if (ebitmap_union(&base_role->roles, &decl_role->roles)) {
-		ERR(state->handle, "Out of memory!");
-		return -1;
-	}
-
-	return 0;
-}
-
-static int populate_roleattributes(link_state_t *state, policydb_t *pol)
-{
-	avrule_block_t *block;
-	avrule_decl_t *decl;
-
-	if (state->verbose)
-		INFO(state->handle, "Populating role-attribute relationship "
-			    "from enabled declarations' local symtab.");
-
-	/* Iterate through all of the blocks skipping the first(which is the
-	 * global block, is required to be present and can't have an else).
-	 * If the block is disabled or not having an enabled decl, skip it.
-	 */
-	for (block = pol->global->next; block != NULL; block = block->next)
-	{
-		decl = block->enabled;
-		if (decl == NULL || decl->enabled == 0)
-			continue;
-
-		if (hashtab_map(decl->symtab[SYM_ROLES].table, 
-				populate_decl_roleattributes, state))
-			return -1;
-	}
-
-	return 0;
-}
-
 /* Link a set of modules into a base module. This process is somewhat
  * similar to an actual compiler: it requires a set of order dependent
  * steps.  The base and every module must have been indexed prior to
@@ -2558,21 +2467,6 @@ int link_modules(sepol_handle_t * handle,
 		goto cleanup;
 	}
 
-	/* Now that all role attribute's roles ebitmap have been settled,
-	 * escalate sub role attribute's roles ebitmap into that of parent.
-	 *
-	 * First, since some role-attribute relationships could be recorded
-	 * in some decl's local symtab(see get_local_role()), we need to
-	 * populate them up to the base.p_roles table. */
-	if (populate_roleattributes(&state, state.base)) {
-		retval = SEPOL_EREQ;
-		goto cleanup;
-	}
-	
-	/* Now do the escalation. */
-	if (hashtab_map(state.base->p_roles.table, expand_role_attributes,
-			&state))
-		goto cleanup;
 
 	retval = 0;
       cleanup:
