@@ -516,11 +516,14 @@ static int is_id_in_scope(struct policydb *pdb, struct stack *decl_stack, char *
 	return is_id_in_scope_with_start(pdb, decl_stack, start, symbol_type, type);
 }
 
-static int semantic_level_to_cil(struct policydb *pdb, int sens_offset, struct mls_semantic_level *level)
+static int semantic_level_to_cil(struct policydb *pdb, struct mls_semantic_level *level)
 {
 	struct mls_semantic_cat *cat;
 
-	cil_printf("(%s ", pdb->p_sens_val_to_name[level->sens - sens_offset]);
+	if (level->sens == 0)
+		return -1;
+
+	cil_printf("(%s ", pdb->p_sens_val_to_name[level->sens - 1]);
 
 	if (level->cat != NULL) {
 		cil_printf("(");
@@ -1586,14 +1589,14 @@ static int range_trans_to_cil(int indent, struct policydb *pdb, struct range_tra
 
 					cil_printf("(");
 
-					rc = semantic_level_to_cil(pdb, 1, &rule->trange.level[0]);
+					rc = semantic_level_to_cil(pdb, &rule->trange.level[0]);
 					if (rc != 0) {
 						goto exit;
 					}
 
 					cil_printf(" ");
 
-					rc = semantic_level_to_cil(pdb, 1, &rule->trange.level[1]);
+					rc = semantic_level_to_cil(pdb, &rule->trange.level[1]);
 					if (rc != 0) {
 						goto exit;
 					}
@@ -2320,7 +2323,7 @@ exit:
 	return rc;
 }
 
-static int user_to_cil(int indent, struct policydb *pdb, struct avrule_block *block, struct stack *UNUSED(decl_stack), char *key, void *datum,  int scope)
+static int user_to_cil(int indent, struct policydb *pdb, struct avrule_block *UNUSED(block), struct stack *UNUSED(decl_stack), char *key, void *datum,  int scope)
 {
 	struct user_datum *user = datum;
 	struct ebitmap roles = user->roles.roles;
@@ -2328,29 +2331,30 @@ static int user_to_cil(int indent, struct policydb *pdb, struct avrule_block *bl
 	struct mls_semantic_range range = user->range;
 	struct ebitmap_node *node;
 	uint32_t i;
-	int sens_offset = 1;
 
-	if (scope == SCOPE_DECL) {
-		cil_println(indent, "(user %s)", key);
-		// object_r is implicit in checkmodule, but not with CIL, create it
-		// as part of base
-		cil_println(indent, "(userrole %s " DEFAULT_OBJECT ")", key);
+	if (scope == SCOPE_REQ) {
+		// if a user is in the REQ scope, then it could cause an
+		// optional block to fail, even if it is never used. However in CIL,
+		// symbols must be used in order to cause an optional block to fail. So
+		// for symbols in the REQ scope, add them to a userattribute as a way
+		// to 'use' them in the optional without affecting the resulting policy.
+		cil_println(indent, "(userattributeset " GEN_REQUIRE_ATTR " %s)", key);
+		return 0;
 	}
 
+	cil_println(indent, "(user %s)", key);
+	// object_r is implicit in checkmodule, but not with CIL, create it
+	// as part of base
+	cil_println(indent, "(userrole %s " DEFAULT_OBJECT ")", key);
+
 	ebitmap_for_each_positive_bit(&roles, node, i) {
 		cil_println(indent, "(userrole %s %s)", key, pdb->p_role_val_to_name[i]);
 	}
 
-	if (block->flags & AVRULE_OPTIONAL) {
-		// sensitivities in user statements in optionals do not have the
-		// standard -1 offset
-		sens_offset = 0;
-	}
-
 	cil_indent(indent);
 	cil_printf("(userlevel %s ", key);
 	if (pdb->mls) {
-		semantic_level_to_cil(pdb, sens_offset, &level);
+		semantic_level_to_cil(pdb, &level);
 	} else {
 		cil_printf(DEFAULT_LEVEL);
 	}
@@ -2359,9 +2363,9 @@ static int user_to_cil(int indent, struct policydb *pdb, struct avrule_block *bl
 	cil_indent(indent);
 	cil_printf("(userrange %s (", key);
 	if (pdb->mls) {
-		semantic_level_to_cil(pdb, sens_offset, &range.level[0]);
+		semantic_level_to_cil(pdb, &range.level[0]);
 		cil_printf(" ");
-		semantic_level_to_cil(pdb, sens_offset, &range.level[1]);
+		semantic_level_to_cil(pdb, &range.level[1]);
 	} else {
 		cil_printf(DEFAULT_LEVEL " " DEFAULT_LEVEL);
 	}
@@ -2506,6 +2510,9 @@ static int level_to_cil(struct policydb *pdb, struct mls_level *level)
 {
 	struct ebitmap *map = &level->cat;
 
+	if (level->sens == 0)
+		return -1;
+
 	cil_printf("(%s", pdb->p_sens_val_to_name[level->sens - 1]);
 
 	if (!ebitmap_is_empty(map)) {
@@ -2544,71 +2551,71 @@ static int context_to_cil(struct policydb *pdb, struct context_struct *con)
 static int ocontext_isid_to_cil(struct policydb *pdb, const char *const *sid_to_string,
 				unsigned num_sids, struct ocontext *isids)
 {
-	int rc = -1;
-
 	struct ocontext *isid;
-
-	struct sid_item {
-		char *sid_key;
-		struct sid_item *next;
-	};
-
-	struct sid_item *head = NULL;
-	struct sid_item *item = NULL;
+	struct ocontext **isid_array;
+	struct strs *strs;
 	char *sid;
-	char unknown[18];
+	char *prev;
 	unsigned i;
 
+	strs = isids_to_strs(sid_to_string, num_sids, isids);
+	if (!strs) {
+		ERR(NULL, "Error writing sid rules to CIL");
+		return -1;
+	}
+
+	if (strs_num_items(strs) == 0) {
+		strs_destroy(&strs);
+		return 0;
+	}
+
+	for (i=1; i < strs_num_items(strs); i++) {
+		sid = strs_read_at_index(strs, i);
+		cil_printf("(sid %s)\n", sid);
+	}
+
+	cil_printf("(sidorder (");
+	prev = NULL;
+	for (i=1; i < strs_num_items(strs); i++) {
+		sid = strs_read_at_index(strs, i);
+		if (prev) {
+			cil_printf("%s ", prev);
+		}
+		prev = sid;
+	}
+	if (prev) {
+		cil_printf("%s", prev);
+	}
+	cil_printf("))\n");
+
+	isid_array = calloc(strs_num_items(strs), sizeof(struct ocontext *));
+	if (!isid_array) {
+		ERR(NULL, "Out of memory");
+		strs_free_all(strs);
+		strs_destroy(&strs);
+		return -1;
+	}
 	for (isid = isids; isid != NULL; isid = isid->next) {
 		i = isid->sid[0];
-		if (i < num_sids && sid_to_string[i]) {
-			sid = (char*)sid_to_string[i];
-		} else {
-			snprintf(unknown, 18, "%s%u", "UNKNOWN", i);
-			sid = unknown;
+		if (i < strs_num_items(strs)) {
+			isid_array[i] = isid;
 		}
-		cil_println(0, "(sid %s)", sid);
-		cil_printf("(sidcontext %s ", sid);
-		context_to_cil(pdb, &isid->context[0]);
-		cil_printf(")\n");
-
-		// get the sid names in the correct order (reverse from the isids
-		// ocontext) for sidorder statement
-		item = malloc(sizeof(*item));
-		if (item == NULL) {
-			ERR(NULL, "Out of memory");
-			rc = -1;
-			goto exit;
-		}
-		item->sid_key = strdup(sid);
-		if (!item->sid_key) {
-			ERR(NULL, "Out of memory");
-			free(item);
-			rc = -1;
-			goto exit;
-		}
-		item->next = head;
-		head = item;
 	}
-
-	if (head != NULL) {
-		cil_printf("(sidorder (");
-		for (item = head; item != NULL; item = item->next) {
-			cil_printf("%s ", item->sid_key);
+	for (i=1; i < strs_num_items(strs); i++) {
+		if (isid_array[i]) {
+			sid = strs_read_at_index(strs, i);
+			cil_printf("(sidcontext %s ", sid);
+			isid = isid_array[i];
+			context_to_cil(pdb, &isid->context[0]);
+			cil_printf(")\n");
 		}
-		cil_printf("))\n");
 	}
+	free(isid_array);
 
-	rc = 0;
+	strs_free_all(strs);
+	strs_destroy(&strs);
 
-exit:
-	while(head) {
-		item = head;
-		head = item->next;
-		free(item->sid_key);
-		free(item);
-	}
-	return rc;
+	return 0;
 }
 
 static int ocontext_selinux_isid_to_cil(struct policydb *pdb, struct ocontext *isids)
@@ -2987,10 +2994,22 @@ static int genfscon_to_cil(struct policydb *pdb)
 	struct genfs *genfs;
 	struct ocontext *ocon;
 	uint32_t sclass;
+	char *name;
+	int wildcard = ebitmap_get_bit(&pdb->policycaps, POLICYDB_CAP_GENFS_SECLABEL_WILDCARD);
+	size_t name_len;
 
 	for (genfs = pdb->genfs; genfs != NULL; genfs = genfs->next) {
 		for (ocon = genfs->head; ocon != NULL; ocon = ocon->next) {
 			sclass = ocon->v.sclass;
+			name = ocon->u.name;
+			name_len = strlen(name);
+			if (wildcard) {
+				if (name_len == 0 || name[name_len - 1] != '*') {
+					ERR(NULL, "genfscon path must end with '*' when genfs_seclabel_wildcard");
+					return -1;
+				}
+				--name_len;
+			}
 			if (sclass) {
 				const char *file_type;
 				const char *class_name = pdb->p_class_val_to_name[sclass-1];
@@ -3011,9 +3030,10 @@ static int genfscon_to_cil(struct policydb *pdb)
 				} else {
 					return -1;
 				}
-				cil_printf("(genfscon %s \"%s\" %s ", genfs->fstype, ocon->u.name, file_type);
+				cil_printf("(genfscon %s \"%.*s\" %s ", genfs->fstype, (int)name_len, name,
+				           file_type);
 			} else {
-				cil_printf("(genfscon %s \"%s\" ", genfs->fstype, ocon->u.name);
+				cil_printf("(genfscon %s \"%.*s\" ", genfs->fstype, (int)name_len, name);
 			}
 			context_to_cil(pdb, &ocon->context[0]);
 			cil_printf(")\n");
